#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>


#define START_STRIKETHROUGH "\e[9m"
#define END_STRIKETHROUGH "\e[29m"

/* Subtract one as sizeof includes the null byte */
#define START_STRIKETHROUGH_LENGTH (sizeof(START_STRIKETHROUGH)-1)
#define END_STRIKETHROUGH_LENGTH (sizeof(END_STRIKETHROUGH)-1)


#define MAX_PRIORITY_STRSIZE 10 /* len(str(2**32)) */
#define MAX_TITLE_SIZE 0x20
#define MAX_TODOS 0x30

#define LIST_OUTPUT_SIZE_PER_TODO (MAX_TITLE_SIZE + MAX_PRIORITY_STRSIZE + START_STRIKETHROUGH_LENGTH + END_STRIKETHROUGH_LENGTH + 3 /* ": \n" */)


typedef struct {
	bool done;
	int32_t priority;
	char title[MAX_TITLE_SIZE+1];
} todo_t;

typedef struct {
	todo_t* storage;
	size_t count;
	size_t capacity;
} todo_manager_t;


todo_manager_t todos = {
	.storage = NULL,
	.count = 0,
	.capacity = 0
};




/*
 *************************************************************************************************
 *************************************** BEGIN SAFE BLOCK ****************************************
 *           To save you some time in the qualification, the functions below are marked
 *      as safe, i.e. without bugs and vulnerabilities. Note, this does not mean that all other
 *                         functions contain bugs or vulnerabilities
 *************************************************************************************************
 *************************************************************************************************
 */



void init(void)
{
	int fd;
	unsigned int rand_val;
	// setting stdout and stdin to be unbuffered
	setvbuf(stdout,0,_IONBF,0);
	setvbuf(stdin,0,_IONBF,0);
	// seeding the random number generator
	fd = open("/dev/urandom", O_RDONLY);
	read(fd, &rand_val, sizeof(rand_val));
	close(fd);
	srand(rand_val);
}


todo_t* todo_manager_find_insertion_point(todo_manager_t* manager, int32_t priority)
{
	size_t count = manager->count;
	todo_t* storage = manager->storage;
	for(size_t i = 0; i < count; ++i)
	{
		if(storage[i].priority < priority)
			return &storage[i];
	}

	return &storage[count];
}

todo_t* todo_manager_find_todo(todo_manager_t* manager, const char* title)
{
	size_t count = manager->count;
	todo_t* storage = manager->storage;
	for(size_t i = 0; i < count; ++i)
	{
		if(!strcmp(title, storage[i].title))
			return &storage[i];
	}
	return NULL;
}


todo_t* todo_manager_add_todo(todo_manager_t* manager, int32_t priority)
{
	if(manager->count >= manager->capacity)
	{
		size_t new_capacity = (manager->capacity * 2) + 1;
		if(new_capacity >= SIZE_MAX / sizeof(todo_t))
			return NULL;

		todo_t* new_storage = realloc(manager->storage, new_capacity * sizeof(todo_t));

		if(!new_storage)
			return NULL;

		manager->storage = new_storage;
		manager->capacity = new_capacity;
	}

	todo_t* insertion_point = todo_manager_find_insertion_point(manager, priority);

	/* Make space via memmove */
	size_t shift_amount = &manager->storage[manager->count] - insertion_point;
	if(shift_amount)
		memmove(insertion_point + 1, insertion_point, shift_amount * sizeof(todo_t));

	manager->count += 1;

	insertion_point->priority = priority;
	return insertion_point;
}

void todo_manager_delete_todo(todo_manager_t* manager, todo_t* todo)
{
	if(!todo)
		return;

	size_t idx = todo - manager->storage;
	size_t count = manager->count;
	size_t capacity = manager->capacity;

	count -= 1;

	size_t shift_amount = count - idx;

	if(shift_amount)
		memmove(todo, todo + 1, shift_amount * sizeof(todo_t));

	manager->count = count;

	if(count <= (capacity / 4))
	{
		size_t new_capacity = count ? (capacity / 2) : 0;

		if(!new_capacity)
		{
			manager->storage = NULL;
			manager->capacity = 0;
		}
		else
		{
			todo_t* new_storage = realloc(manager->storage, new_capacity * sizeof(todo_t));

			if(!new_storage)
				return;

			manager->storage = new_storage;
			manager->capacity = new_capacity;
		}
	}
}


void todo_manager_resort_todo(todo_manager_t* manager, todo_t* todo)
{
	size_t count = manager->count;
	if (count == 1)
		return;
	todo_t* storage = manager->storage;
	size_t idx = todo - storage;

	ssize_t end;
	ssize_t direction = 0;


	if(idx != 0)
	{
		if(todo[-1].priority < todo[0].priority)
		{
			direction = -1;
			end = 0;
		}
	}

	if(!direction && idx != (count - 1))
	{
		if(todo[1].priority > todo[0].priority)
		{
			direction = 1;
			end = count-1;
		}
	}

	if(!direction)
		return;

	todo_t tmp = *todo;

	ssize_t i;
	for(i = idx; i != end; i += direction)
	{
		ssize_t diff = storage[i+direction].priority - tmp.priority;
		if(!diff || (diff > 0) != (direction > 0))
			break;
      storage[i] = storage[i+direction];
	}
	storage[i] = tmp;
}

/*
 *************************************************************************************************
 **************************************** END SAFE BLOCK *****************************************
 *           To save you some time in the qualification, the functions above are marked
 *      as safe, i.e. without bugs and vulnerabilities. Note, this does not mean that all other
 *                         functions contain bugs or vulnerabilities.
 *************************************************************************************************
 *************************************************************************************************
 */












void win(void)
{
    execl("/bin/get_flag", "get_flag", NULL);
}

bool read_title(char* buf, size_t buf_size)
{
	if(!fgets(buf, buf_size, stdin))
	{
		printf("Could not read title...\n");
		return false;
	}

	if(buf[0] == '\n')
	{
		printf("Title cannot be empty!\n");
		return false;
	}

	/* Remove newline in title */
	strtok(buf, "\n");

	if(strlen(buf) > MAX_TITLE_SIZE)
	{
		printf("Title too long!");
		return false;
	}
	return true;
}



bool read_priority(char* priority_buf, size_t buf_size, int32_t* ret)
{
	printf("Priority: ");
	if(!fgets(priority_buf, buf_size, stdin))
	{
		printf("Could not read priority...\n");
		return false;
	}

	if(sscanf(priority_buf, "%d", ret) != 1)
	{
		printf("Invalid input!");
		return false;
	}
	return true;
}





void add_todo(void)
{
	int32_t priority;
	/* +2 for newline + null-byte */
	char title[MAX_TITLE_SIZE+2];
	char priority_buf[MAX_PRIORITY_STRSIZE+2];

	if(todos.count >= MAX_TODOS)
	{
		printf("Maximum amount of todos for this list reached!\n");
		return;
	}

	if(!read_priority(priority_buf, sizeof(priority_buf), &priority))
		return;

	printf("Title: ");
	if(!read_title(title, sizeof(title)))
		return;

	if(todo_manager_find_todo(&todos, title))
	{
		printf("ToDo already exists!\n");
		return;
	}

	todo_t* todo = todo_manager_add_todo(&todos, priority);
	if(!todo)
	{
		printf("Out of memory!\n");
		return;
	}
	todo->done = false;
	strcpy(todo->title, title);
	printf("ToDo '%s' added!\n", title);
}

void complete_todo(void)
{
	/* +2 for newline + null-byte */
	char title[MAX_TITLE_SIZE+2];

	printf("ToDo to mark as completed: ");
	if(!read_title(title, sizeof(title)))
		return;

	todo_t* todo = todo_manager_find_todo(&todos, title);

	if(!todo)
	{
		printf("ToDo doesn't exist!\n");
		return;
	}

	if(todo->done)
		printf("ToDo already marked as completed!\n");
	else {
		todo->done = true;
		printf("ToDo '%s' marked as completed!\n", title);
	}
}


void delete_todo(void)
{
	/* +2 for newline + null-byte */
	char title[MAX_TITLE_SIZE+2];

	printf("ToDo to delete: ");
	if(!read_title(title, sizeof(title)))
		return;

	todo_t* todo = todo_manager_find_todo(&todos, title);

	if(!todo)
	{
		printf("ToDo doesn't exist!\n");
		return;
	}
	todo_manager_delete_todo(&todos, todo);
}



void edit_todo_title(void)
{
	/* +2 for newline + null-byte */
	char title[MAX_TITLE_SIZE+2];

	printf("Old title: ");
	if(!read_title(title, sizeof(title)))
		return;

	todo_t* todo = todo_manager_find_todo(&todos, title);

	if(!todo)
	{
		printf("ToDo doesn't exist!\n");
		return;
	}

	printf("New title: ");
	if(!read_title(title, sizeof(title)))
		return;

	strcpy(todo->title, title);
}

void edit_todo_priority(void)
{
	int32_t priority;
	/* +2 for newline + null-byte */
	char priority_buf[MAX_PRIORITY_STRSIZE+2];
	char title[MAX_TITLE_SIZE+2];

	printf("ToDo to modify: ");
	if(!read_title(title, sizeof(title)))
		return;

	todo_t* todo = todo_manager_find_todo(&todos, title);

	if(!todo)
	{
		printf("ToDo doesn't exist!\n");
		return;
	}

	printf("Priority: ");
	if(!read_priority(priority_buf, sizeof(priority_buf), &priority))
		return;

	todo->priority = priority;
	todo_manager_resort_todo(&todos, todo);
}


void list_todos(void)
{
	if(!todos.count)
	{
		printf("No ToDos!\n");
		return;
	}
	/* Optimization: Allocating on stack is fast, we can always alloc max size*/
	char outputBuf[LIST_OUTPUT_SIZE_PER_TODO * MAX_TODOS + 1 /* NULL byte */];

	/* Optimization: Prepare whole buffer to reduce number of syscalls */
	char* loc = outputBuf;

	for(size_t i = 0, count = todos.count; i < count; ++i)
	{
		if(i) *(loc++) = '\n';
		todo_t* todo = &todos.storage[i];

		/* snprintf includes NULL-byte in size calculation, but
		 * not in return value, thus add + 1 for n */
		loc += snprintf(loc, LIST_OUTPUT_SIZE_PER_TODO + 1,
						"%s%d: %s%s",
						todo->done ? START_STRIKETHROUGH : "",
						todo->priority,
						todo->title,
						todo->done ? END_STRIKETHROUGH : "");
	}
	puts(outputBuf);
}




void print_welcome(void)
{
	printf("Welcome to ToDo-List-as-a-Service Demo!\n"
			 "Manage ToDo-List to your liking with a simple terminal interface!\n");
}


void print_menu(void)
{
	puts("=================================");
	puts(" What would you like to do next?");
	puts(" A: Add ToDo");
	puts(" C: Mark ToDo as completed");
	puts(" D: Delete ToDo");
	puts(" E: Edit ToDo Title");
	puts(" I: Increase Priority");
	puts(" L: List ToDos");
	puts(" P: Edit ToDo Priority");
	puts(" Q: Quit");
	puts("=================================");
	printf("> ");
}


int main()
{
	char choice[0x20];
	init();
	print_welcome();
	while (1)
	{
		print_menu();
		if(!fgets(choice, sizeof(choice), stdin))
			return 0;

		switch(choice[0])
		{
			case 'a':
			case 'A':
				add_todo();
				break;
			case 'c':
			case 'C':
				complete_todo();
				break;
			case 'd':
			case 'D':
				delete_todo();
				break;
			case 'e':
			case 'E':
				edit_todo_title();
				break;
			case 'l':
			case 'L':
				list_todos();
				break;
			case 'p':
			case 'P':
				edit_todo_priority();
				break;
			case 'q':
			case 'Q':
				exit(0);
		}
	}

}
